function [segs, seg_lambdas, seg_to_seed_mapping, mincut_vals, t_pmc, ...
          cut_info] = multiseed_param_min_st_cut(gp_obj, method_str, ...
                        graph_start_idxs, graphcut_params)
% options: 
%   nodynamic*: recreate the graph for every seed and parameter but use TBB 
%               parallelization to simultaneously solve different seeds
%
%   kohli_reusetrees*: create graph once for each seed and reparameterize 
%               the graph (Kohli PAMI 2007) for every seed and use TBB 
%               parallelization to solve seeds simultaneously
%
%   kohli*: create graph once but don't reuse trees
%
%   multiseed*: Uses the graph reuse scheme over multiple seeds. Given
%               multiple seeds it first generates a precomputation graph. 
%               That graph is reparameterized for each seed and solved 
%               consequently for each seed separately
%
%   multiseed_opplambda*: Uses same scheme as multiseed but the
%               precomputation graph is built from the opposite end of the 
%               first scheduled lambda (so if the first lambda is scheduled 
%               is at index 1, then the lambda used to build the 
%               precomputation graph would be the lambda at the last index)
%
% Both kohli and multiseed option can be appended with the *_rev* option.
% This schedules the parametric min-cut in the reverse direction (i.e.
% descending from the largest lambda), which might give some speed benefit
% if the first cut is just the seed in the first few parameters - but on
% the other hand might be disadvantageous if the cut for the last few
% lambdas was the full graph on the src side.
%
% (Every option can be appended with '_noparallel' to stop parallel
% computation of seeds)
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Fall 2014

    % get the logging function
    rfprintf = gp_obj.seg_obj.other_params.printf_func;

    % generate the options string for each graph type
    options_str = '';
    
    REV = false(1, length(graph_start_idxs));
    
    for graph_type_idx = 1:length(graph_start_idxs)
        options_str = [options_str, '|', method_str];
        
        % if the graph-cut method is Kohli&Torr or ours and was the lambda
        % scheduling needs to be set in reverse
        if ~isempty(graphcut_params) && ...
           (~isempty(strfind(gp_obj.segm_params.pmc_maxflow_method, 'kohli')) || ...
            ~isempty(strfind(gp_obj.segm_params.pmc_maxflow_method, 'multiseed'))) ...
            && graphcut_params(graph_type_idx) == 0
            REV(graph_type_idx) = true;
        end

        if REV(graph_type_idx)
            options_str = [options_str '_rev'];
        end
    end
    options_str(1) = [];

    % adjust graph (like thresholding values and data conversion) before
    % creating the final data structure used by graph cuts
    [nonlambda_s, nonlambda_t, lambda_s, lambda_t, lambda_range, ...
        pairwise_edges, DISC_FACTOR, BIG_VALUE] = ...
            preprocess_graphs(gp_obj, ...
                              gp_obj.graph_unaries_all.nonlambda_s, ...
                              gp_obj.graph_unaries_all.nonlambda_t, ...
                              gp_obj.graph_unaries_all.lambda_s, ...
                              gp_obj.graph_unaries_all.lambda_t);
    
    t_pmc = tic;
    
    % compute maxflow via boykov-kolmogorov
    [all_cuts, meta_info] = ...
        bk_dynamicgraphs_mex(nonlambda_s, nonlambda_t, lambda_s, ...
                             lambda_t, lambda_range, pairwise_edges, ...
                             graph_start_idxs, options_str, true);
    
    t_pmc = toc(t_pmc);
    t_pmc = repmat(t_pmc / length(graph_start_idxs), ...
                   length(graph_start_idxs), 1);
               
    rfprintf(' ... ');
    
    % notes the starting seed index of each graph type
    graph_start_idxs2 = [graph_start_idxs, size(nonlambda_s,2)+1];
    
    % initialize variables
    segs = false(size(nonlambda_s,1), 0);
    last_cut_idx = 0;
    seg_to_cut_mapping = [];
    all_fg_seg = false(1,size(all_cuts,2));
    seg_lambdas = [];
    mincut_vals = NaN;
    
    % store the number segments generated for each seed
    num_segs_per_seed = zeros(1,size(all_cuts,2));
    
    % Iterate over all seed indices, and extract segments produced from 
    % each seed. Also generate the set of lambdas associated with each
    % segment. Also keep a note of the mapping of each segment to a cut.
    for seed_idx = 1:size(all_cuts,2)
        % find what graphy type, this set of parametric cuts belong to
        curr_graph_type_idx = find(seed_idx < graph_start_idxs2, 1, 'first') - 1;
        
        % since parametric cut strictly grow (or if in reverse, strictly
        % shrinks), the vector of values here tell at which lambda 1+ index
        % they became part of the src cut (or sink cut, when run in
        % reverse). If a variable is zero it means that the variable always 
        % remained in the sink side (and if in reverse, then the variable 
        % always remained in the src side). Lets say there are 4 parameters 
        % (lambdas), e.g. cuts:
        % forward: [1 3 4 0] -> 0 = always sink
        % reverse: [0 3 4 4] -> 0 = always src
        curr_cuts = all_cuts(:,seed_idx);
        
        if REV(curr_graph_type_idx)
            % in reverse variables which are 0 are the variables that 
            % always remained src side of the cut
            
            % converting to cuts produced in the forward direction:
            % consider a variable whose label changes with increasing
            % lambda as follows:
            %       0 0 0 0 1 1 1 1
            % If parametric min-cut is computed with increasing (fwd) 
            % lambdas, we will get the lambda index 5 for this variable. If
            % parametric min-cut is computed with decreasing (reverse)
            % lambdas, we will get the lambda index 4. Hence we just need
            % to do the following to get lambda index equivalent to
            % increasing lambda scheduling
            curr_cuts = curr_cuts + 1;
            curr_cuts(curr_cuts > length(lambda_range)) = 0;
        end
        
        % if a variable is 0, it means it always remained in the sink side
        always_sink = (curr_cuts == 0);
        
        % get all the breakpoints (value of 0 means that it never got
        % converted to src)
        bp = double(unique(curr_cuts))';
        bp(bp == 0) = [];
        output_segs = bsxfun(@le, curr_cuts, bp);
        output_segs(always_sink,:) = 0;
        
        % get the lambdas for each break-point
        curr_lambdas = lambda_range(bp);
        
        % check if last segment is all fg (then discard) [we cant get all
        % bg solutions because there is an infinity cost at fg seeds].
        if ~isempty(output_segs) && all(output_segs(:,end))
            output_segs(:,end) = [];
            curr_lambdas(end) = [];
            bp(end) = [];
            all_fg_seg(seed_idx) = true;
        end
        
        % count the number of cuts produced for this seed
        curr_cut_locs = (meta_info(7,:) == seed_idx-1);
        num_cuts = nnz(curr_cut_locs);
        
        % compute the mapping from the segments produced to the cuts
        if REV(curr_graph_type_idx)
            % since the lambda iterates in descending order, and since we
            % transformed the breakpoints previously, we need the following
            % arithmetic to get the cuts associated to the sets
            lambda_iter = (length(lambda_range) - bp) + 2;
            lambda_iter(lambda_iter > num_cuts) = num_cuts;
            curr_seg_to_cut_mapping = last_cut_idx + lambda_iter;
        else
            curr_seg_to_cut_mapping = last_cut_idx + bp;
        end
        
        % collate segments, if there are any segments for this seed
        if ~isempty(output_segs)
            seg_lambdas = [seg_lambdas, single(curr_lambdas)];
            seg_to_cut_mapping = [seg_to_cut_mapping, ...
                                  uint16(curr_seg_to_cut_mapping)];
            segs = [segs, output_segs];
            num_segs_per_seed(seed_idx) = size(output_segs,2);
        end
        
        last_cut_idx = last_cut_idx + num_cuts;
    end
    
    % sanity check results
    assert(last_cut_idx == size(meta_info,2), ['The number of cuts ', ...
           'counted is not consistent with graph cut results']);
    assert(length(seg_to_cut_mapping) == size(segs,2), ['The segments ' ...
           'to cut mapping size not equal to number of segments']);
    assert(~any(seg_to_cut_mapping > size(meta_info,2)), ['Some ' ...
           'semgent to cut mapping is greater than number of cuts']);
    
    % get the seed index for each cut
    cuts_seed_idxs = uint16(meta_info(7,:) + 1);
    % get the graph type index for each cut
    cuts_graph_type_idxs = ...
        uint8(arrayfun(@(x) find(x<graph_start_idxs2, 1, 'first') - 1, ...
              cuts_seed_idxs));
    % get the seed mapping for all segments
    seg_to_seed_mapping = cuts_seed_idxs(seg_to_cut_mapping);
    
    % remaining sanity checks
    assert(all(cuts_graph_type_idxs > 0 & ...
               cuts_graph_type_idxs <= length(graph_start_idxs)), ...
           'Some graph type indices for cuts are out of bound');
    
    % store results
    cut_info.gc_time_all = meta_info(1:2,:);
    cut_info.gc_nsteps_all = uint16(meta_info(3:5,:));
    cut_info.gc_flowvals_all = meta_info(6,:);
    cut_info.cuts_graph_type_idxs = cuts_graph_type_idxs;
    cut_info.cuts_seed_idxs = cuts_seed_idxs;
    cut_info.seg_to_cut_mapping = seg_to_cut_mapping;
    cut_info.all_fg_seg = all_fg_seg;
end
