classdef UniformGraph < AbstractGraph
%UNIFORMGRAPH is a class which generates graphs with seeds and unaries
% which are not dependent on image information (for instance unaries are
% only based on superpixel size). This class is derived from AbstractGraph. 
% To undestand how unaries are generated, its worth seeing how 
% AbstractGraph.prepare_graphs() calls create_unaries() in this class to
% compute unary values.
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Summer 2014

    properties
    end
    
    methods
        function obj = UniformGraph(seg_obj, graph_idx)
            obj.init(seg_obj, graph_idx);
        end
        
        function create_unaries(obj, graph_sub_method, seed_sets)
            switch graph_sub_method
                case 'internal'
                    obj.generate_uniform_unaries('internal', seed_sets);
                case 'external'
                    obj.generate_uniform_unaries('external', seed_sets);
                case 'subframe'
                    error('UniformGraph:compute_graph', ...
                          '''subframe'' not implemented yet');
                otherwise
                    error('UniformGraph:compute_graph', ...
                          '''%s'' is an invalid graph method', ...
                          graph_sub_method);
            end
        end
        
        function generate_uniform_unaries(obj, internal_external, ...
                                          seed_sets)
            not_border = ~obj.seg_obj.sp_frame_set;
            
            funcs = create_unary_aux_funcs(obj, internal_external);
            
            % find the offset according to the number of unary graphs
            % already added to graph_unaries_all
            ofst = size(obj.graph_unaries_all.nonlambda_s,2);
            
            % initialize the matrices for holding all the unaries (see the
            % comment for this structure in AbstractGraph.prepare_graphs())
            obj.graph_unaries_all.nonlambda_s = ...
                [obj.graph_unaries_all.nonlambda_s, zeros(size(seed_sets))];
            obj.graph_unaries_all.nonlambda_t = ...
                [obj.graph_unaries_all.nonlambda_t, zeros(size(seed_sets))];
            obj.graph_unaries_all.lambda_s = ...
                [obj.graph_unaries_all.lambda_s, zeros(size(seed_sets))];
            obj.graph_unaries_all.lambda_t = ...
                [obj.graph_unaries_all.lambda_t, zeros(size(seed_sets))];
            
            % set the seed mapping array (see the comment for this
            % structure in AbstractGraph.prepare_graphs()). We can directly
            % set this because each seed is going to generate a set of
            % unaries (each seed would generate a graph)
            obj.graph_unaries_seed_mapping = ...
                [obj.graph_unaries_seed_mapping, 1:size(seed_sets, 2)];
            
            % iterate over all seed locations
            for i = 1:size(seed_sets, 2)
                if strcmp(internal_external, 'internal')
                    [nonlambda_s, lambda_t] = internal_uniform(obj, ...
                                                     seed_sets(:,i), ...
                                                     funcs.parametric_func);
                    
                    obj.graph_unaries_all.nonlambda_s(:,ofst+i) = ...
                                                               nonlambda_s;
                    obj.graph_unaries_all.lambda_t(:,ofst+i) = lambda_t;
                else
                    [nonlambda_s, nonlambda_t, lambda_s, lambda_t] = ...
                        external_uniform(obj, seed_sets(:,i), ...
                                         not_border, ...
                                         funcs.parametric_func, ...
                                         funcs.frame_cost_func);
                    
                    obj.graph_unaries_all.nonlambda_s(:,ofst+i) = ...
                                                               nonlambda_s;
                    obj.graph_unaries_all.nonlambda_t(:,ofst+i) = ...
                                                               nonlambda_t;
                    obj.graph_unaries_all.lambda_s(:,ofst+i) = lambda_s;
                    obj.graph_unaries_all.lambda_t(:,ofst+i) = lambda_t;
                end
            end
        end
        
        function [nonlambda_s, lambda_t] = internal_uniform(obj, ...
                                                fg_seed, parametric_func)
            % see the comment for these output matrices in 
            % AbstractGraph.prepare_graphs()
            nonlambda_s = zeros(length(fg_seed),1);
            lambda_t = zeros(length(fg_seed),1);
            
            not_seed = ~fg_seed;
            nonlambda_s(fg_seed) = inf;
            lambda_t(not_seed) = parametric_func(not_seed);
        end
        
        function [nonlambda_s, nonlambda_t, lambda_s, lambda_t] = ...
                external_uniform(obj, fg_seed, not_border, ...
                                 parametric_func, frame_cost_func)
            % see the comment for these output matrices in 
            % AbstractGraph.prepare_graphs()
            nonlambda_s = zeros(length(fg_seed),1);
            nonlambda_t = zeros(length(fg_seed),1);
            lambda_s = zeros(length(fg_seed),1);
            lambda_t = zeros(length(fg_seed),1);
            
            not_brdr_nor_seed = not_border & ~fg_seed;
            lambda_weights = parametric_func(not_brdr_nor_seed);
            bground_pixels = ~(not_border | obj.seg_obj.sp_frame_set);
            
            nonlambda_s(fg_seed) = inf;
            nonlambda_t(bground_pixels) = inf;
            lambda_s(not_brdr_nor_seed) = lambda_weights;
            
            frame_cost = frame_cost_func(fg_seed);
            
            nonlambda_t(obj.seg_obj.sp_frame_set) = ...
                frame_cost * obj.graph_seed_frame_weight;
            lambda_t(not_brdr_nor_seed) = lambda_weights;
        end
    end
end
