function download_algorithms(light_ver, simulate_no_download)
% Downloads and compiles/preps different external codes needed for RIGOR
%
% @input params:
% light_ver: If set to true, it forces the script to only download 
%       libraries absolutely necessary for the default settings of RIGOR.
%
% simulate_no_download: It forces the script to not download anything. Its
%       useful when all the downloads are already present - the script just 
%       needs to be used for changing files (useful for debugging).
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Summer 2014

clc

% if you only want the absolutely necessary files to run RIGOR
if ~exist('light_ver', 'var')
    light_ver = false;
end

% if you don't want to download the files (for testing this script)
if ~exist('simulate_no_download', 'var')
    simulate_no_download = false;
end

no_download_dir = 'pre_downloaded';

fprintf(2, '---------------------------------------------------------------------------\n                              Download Notice                              \n---------------------------------------------------------------------------\nBy running this script, you accept all licensing agreements accompanied\nwith the 3rd party softwares that will now be downloaded and used later in\nour scripts.\n\nPress ''y'' to accept (any other key to stop the script): ');
user_inp = input('','s');
if isempty(user_inp) || ~strncmpi(user_inp(1), 'y', 1)
    return;
end

% destination algorithms directory
curr_dir = pwd;
code_root_dir = fullfile(fileparts(which(mfilename)), '..');
utils_dir = fullfile(code_root_dir, 'utils');
prepkgd_extern_src_rel = 'prepkgd_extern_src';
prepkgd_dir = fullfile(code_root_dir, prepkgd_extern_src_rel);
extern_src_rel = 'extern_src';
algos_dir = fullfile(code_root_dir, extern_src_rel);
if exist(algos_dir,'dir')
    rmdir(algos_dir, 's');
end
mkdir(algos_dir);

svn_basepath = 'https://github.com/ahmadh84/extern_sync/trunk/';

% create temp download directory
if ~simulate_no_download
    temp_dir = fullfile(code_root_dir, sprintf('temp%d', randi(1e8)));
    mkdir(temp_dir);
else
    temp_dir = no_download_dir;
end

try
    % find if 64 bit machine
    is64 = ~isempty(strfind(computer, '64'));
    
    % Point Fuxin library to the right directory
    fuxin_lib_rel = fullfile(prepkgd_extern_src_rel, 'fuxin_lib_src');
    replaceInTextFile(fullfile(code_root_dir, 'internal_params.m'), '''/home/ahumayun/videovolumes/fuxin_lib_src''', ['fullfile(fp.code_root_dir, ''', fuxin_lib_rel, ''')'], true);
    fuxin_lib_rel = fullfile(prepkgd_dir, 'fuxin_lib_src');
    classregtree_dir = fullfile(fuxin_lib_rel, '@classregtree_fuxin', 'private');
    fprintf(1, 'mex''ing fuxin library code\n');
    eval(sprintf('mex -O %s/regtreeEval.cpp -output %s/regtreeEval', classregtree_dir, classregtree_dir));
    eval(sprintf('mex -O %s/regtree_findbestsplit.cpp -output %s/regtree_findbestsplit', classregtree_dir, classregtree_dir));
    
    % compute Fuxin library myqueue
    myqueue_dir = fullfile(fuxin_lib_rel, 'myqueue_1.1');
    eval(sprintf('mex -O %s/pq_create.cpp -output %s/pq_create', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_delete.cpp -output %s/pq_delete', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_demo.cpp -output %s/pq_demo', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_pop.cpp -output %s/pq_pop', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_push.cpp -output %s/pq_push', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_size.cpp -output %s/pq_size', myqueue_dir, myqueue_dir));
    eval(sprintf('mex -O %s/pq_top.cpp -output %s/pq_top', myqueue_dir, myqueue_dir));
    fprintf(1, 'Done mex''ing\n');
    
    % prepare Fuxin library segments functions
    segments_dir = fullfile(fuxin_lib_rel, 'segments');
    eval(sprintf('mex -O %s/find_unique_elems_in_neighborhood.cpp -output %s/find_unique_elems_in_neighborhood', segments_dir, segments_dir));
    
    % compile poly2mask_mex (used for scoring COCO GT)
    openmp_flags = 'CFLAGS="\$CFLAGS -fopenmp" CXXFLAGS="\$CXXFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp" COMPFLAGS="\$COMPFLAGS /openmp"';
    eval(sprintf('mex -O %s/poly2mask_mex.cpp %s -output %s/poly2mask_mex', utils_dir, openmp_flags, utils_dir));
    
    
    % move stein boundary processing code to the segmentation folder
%     mkdir(fullfile(algos_dir, 'segmentation'));
%     copyfile(fullfile(algos_dir, 'stein_boundaryprocessing'), fullfile(algos_dir, 'segmentation', 'stein_boundaryprocessing'));
    success = download_svn([svn_basepath, 'segmentation/stein_boundaryprocessing'], ...
                 fullfile(algos_dir, 'segmentation', 'stein_boundaryprocessing'), ...
                 'Stein boundary processing', simulate_no_download);
    
    
    % download Piotr's toolbox code
    success = download_svn([svn_basepath, 'toolboxes/piotr_toolbox'], ...
                    fullfile(algos_dir, 'toolboxes', 'piotr_toolbox'), ...
                    'Piotr''s toolbox code', simulate_no_download);

    if success
        % do not compile for linux - there is some problem when running
        % struct edges with compiled code (rather than with the compiled
        % code that comes with the library)
        if ~isunix
            cd(fullfile(algos_dir, 'toolboxes', 'piotr_toolbox', 'external'));
            toolboxCompile;
            fprintf(1, 'Done mex''ing\n');
            cd(curr_dir);
        end
    end
    
    
    % download Piotr's Struct Edges code
    success = download_svn([svn_basepath, 'segmentation/boundaries--dollar_ICCV_2013_structedges'], ...
                    fullfile(algos_dir, 'segmentation', 'boundaries--dollar_ICCV_2013_structedges'), ...
                    'Piotr''s Struct Edges code', simulate_no_download);
    
    if success
        cd(fullfile(algos_dir, 'segmentation', 'boundaries--dollar_ICCV_2013_structedges'));
        % use piotr's compiled version (they seem to be faster
        if isunix
            mex -O private/edgesDetectMex.cpp -outdir private '-DUSEOMP' CXXFLAGS="\$CXXFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
            mex -O private/edgesNmsMex.cpp    -outdir private '-DUSEOMP' CXXFLAGS="\$CXXFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
            mex -O private/spDetectMex.cpp    -outdir private '-DUSEOMP' CXXFLAGS="\$CXXFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
            mex -O private/edgeBoxesMex.cpp   -outdir private
        else
            % if windows
            mex -O private/edgesDetectMex.cpp -outdir private '-DUSEOMP' 'OPTIMFLAGS="$OPTIMFLAGS' '/openmp"'
            mex -O private/edgesNmsMex.cpp    -outdir private '-DUSEOMP' 'OPTIMFLAGS="$OPTIMFLAGS' '/openmp"'
            mex -O private/spDetectMex.cpp    -outdir private '-DUSEOMP' 'OPTIMFLAGS="$OPTIMFLAGS' '/openmp"'
            mex -O private/edgeBoxesMex.cpp   -outdir private
        end
        fprintf(1, 'Done mex''ing\n');
        cd(curr_dir);
    end
    
    
    % download export_fig code
    success = download_svn([svn_basepath, 'utils/export_fig'], ...
                    fullfile(algos_dir, 'utils', 'export_fig'), ...
                    'Woodford''s matlab export fig code', simulate_no_download);
    
    if success
        replaceInTextFile(fullfile(utils_dir, 'dFig.m'), '~/extern_src', algos_dir, true);
    end
    
    
    if ~light_ver
        % download Leordeanu''s GB code
        success = download_svn([svn_basepath, 'segmentation/boundaries--leordeanu_ECCV_2012_gb'], ...
                    fullfile(algos_dir, 'segmentation', 'boundaries--leordeanu_ECCV_2012_gb'), ...
                    'Leordeanu''s GB flow code', simulate_no_download);
        success = download_svn([svn_basepath, 'segmentation/segbench/lib/matlab'], ...
                    fullfile(algos_dir, 'segmentation', 'segbench', 'lib', 'matlab'), ...
                    'Berkeley segbench', simulate_no_download);

        
        % download Joseph Lim's SketchTokens code
        success = download_code('https://github.com/joelimlimit/SketchTokens/archive/master.zip', ...
            fullfile(temp_dir, 'master.zip'), 'Joseph Lim''s SketchTokens code', simulate_no_download);
        success = success & download_code('http://people.csail.mit.edu/lim/lzd_cvpr2013/st_data.tgz', ...
            fullfile(temp_dir, 'st_data.tgz'), 'Joseph Lim''s SketchTokens data', simulate_no_download);

        if success
            mkdir(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens'));
            unzip(fullfile(temp_dir, 'master.zip'), fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens'));
            untar(fullfile(temp_dir, 'st_data.tgz'), fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens'));
            rmdir(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens', 'SketchTokens-master', 'models'), 's');
            movefile(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens', 'models'), fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens', 'SketchTokens-master'));
            movefile(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens', 'SketchTokens-master', '*'), fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens'));
            rmdir(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens', 'SketchTokens-master'), 's');
            cd(fullfile(algos_dir, 'segmentation', 'boundaries--lim_CVPR_2013_sketchtokens'));
            mex -O stDetectMex.cpp 'OPTIMFLAGS="$OPTIMFLAGS' '/openmp"'
            fprintf(1, 'Done mex''ing\n');
            cd(curr_dir);
        end
        
        
        % download Woodford's Image-Based Rendering and Stereo code
        success = download_svn([svn_basepath, 'segmentation/imrender'], ...
            fullfile(algos_dir, 'segmentation', 'imrender'), ...
            'Woodford''s Image-Based Rendering and Stereo code', simulate_no_download);

        if success
            cd(fullfile(algos_dir, 'segmentation', 'imrender', 'vgg'));
            % segment as a test - which also compiles the file
            temp = vgg_segment_gb(imread('peppers.png'), 0.5, 10, 10);
            fprintf(1, 'Done mex''ing\n');
            cd(curr_dir);
        end
        
        
        % download Vedaldi's vlfeat code
        success = download_code('https://github.com/vlfeat/vlfeat/archive/master.zip', ...
            fullfile(temp_dir, 'vlfeat.zip'), 'Vedaldi''s vlfeat code', simulate_no_download);
    
        if success
            unzip(fullfile(temp_dir, 'vlfeat.zip'), fullfile(algos_dir, 'toolboxes'));
            d = dir(fullfile(algos_dir, 'toolboxes', '*vlfeat*'));
            movefile(fullfile(algos_dir, 'toolboxes', d(1).name), fullfile(algos_dir, 'toolboxes', 'vlfeat'));
        end
    end
    
    
    % change the extern code path in the internal parameters script
    replaceInTextFile(fullfile(code_root_dir, 'internal_params.m'), '''/home/ahumayun/extern_src''', ['fullfile(fp.code_root_dir, ''', extern_src_rel, ''')'], true);

    
    % delete temp directory
    if ~simulate_no_download
        rmdir(temp_dir, 's');
    end
    
catch exception
    % remove temp dir
    rmdir(temp_dir, 's');
    rethrow(exception)
end

end


function success = download_code(url, filepath, desc, simulate_no_download)
try
    fprintf(1, 'Downloading %s ...\n', desc);
    if ~simulate_no_download
        [f, status] = urlwrite(url, filepath);
        success = status == 1;
    else
        success = true;
    end
catch exception
    success = false;
end

if success
    if ~simulate_no_download
        fprintf(1, 'Done downloading\n');
    else
        fprintf(1, 'Simulated download call\n');
    end
else
    fprintf(2, 'Downloading failure\n');
end
end


function success = download_svn(url, downloadpath, desc, simulate_no_download)
try
    fprintf(1, 'Downloading %s ...\n', desc);
    if ~simulate_no_download
        [status,cmdout] = system(sprintf('svn export "%s" "%s"', url, downloadpath));
        success = status == 0;
    else
        success = true;
    end
catch exception
    success = false;
end

if success
    if ~simulate_no_download
        fprintf(1, 'Done downloading\n');
    else
        fprintf(1, 'Simulated download call\n');
    end
else
    fprintf(2, 'Downloading failure\n');
end
end


function transferMatlabFromGNU(makefilepath, dest_dir)
fd = fopen(makefilepath, 'r');
str = fread(fd);
str = char(str');
fclose(fd);

files = {};
re_matches = regexp(str, '(?:matlab)\s+:=\s+(\S+.(?:(mat)|m)(?:\s*\\?\s*((\$\(wildcard)|(\)))?\s+))*', 'tokens');
for idx = 1:length(re_matches)
    f = regexp(re_matches{idx}{1}, '(\S+.(?:(mat)|m))', 'tokens');
    files = [files cellfun(@(x)x, f)];
end
for idx = 1:length(files)
    movefile(fullfile(fileparts(makefilepath), files{idx}), dest_dir);
end
end


function adjustAttributes(folder_path)
% dont need to change file permissions on windows
if ispc == 1
    return;
end

d = dir(folder_path);
for idx = 1:length(d)
    if strcmp(d(idx).name,'.') || strcmp(d(idx).name,'..')
        continue;
    end
    
    curr_path = fullfile(folder_path,d(idx).name);
    if d(idx).isdir == 1
        adjustAttributes(curr_path);
    else
        unix(['chmod 0644 "' curr_path '"']);
    end
end
end
