function rigor_batch_run(SECTIONS, section_no, data_dir, gt_dir, ...
                         batchlist_filepath, file_ext, analysis_ims_dir, ...
                         do_analysis, per_im_out, dataset_id, varargin)
% You can use rigor_batch_run to automatically run rigor_obj_segments for a 
% batch of a files with the same set of parameter settings. In one call to 
% rigor_batch_run.m, you can either compute segments for all the files 
% specified, or just for a subset of them (useful when parallelizing 
% processing on a cluster). You can also use the function for analyzing 
% results (quantitative scores and timing) of each run of 
% rigor_obj_segments.m. The function works seamlessly for processing images 
% from the PASCAL VOC dataset.
%
% The command is as follows: 
%     rigor_batch_run(SECTIONS, section_no, data_dir, gt_dir, ...
%                     batchlist_filepath, file_ext, analysis_ims_dir, ...
%                     do_analysis, per_im_out, dataset_id, varargin);
%
% For instance, if you wanted to process all the 1449 images of the PASCAL 
% VOC 2012 validation set. I am supposing that the complete dataset is 
% stored at ~/pascalvoc12: 
%     rigor_batch_run(1, 1, '~/pascalvoc12/JPEGImages', ...
%                     '~/pascalvoc12/SegmentationObject', ...
%                     '~/pascalvoc12/ImageSets/Segmentation/val.txt', ...
%                     '.jpg', '', false, false, 'pascalvoc', ...
%                     'force_recompute', true);
% This command would compute object proposals for all the validation images 
% and stores them to disk. The last two parameters forces 
% rigor_obj_segments.m to recompute all segments if some/all had been 
% computed before and stored to disk. You can also ask for generating 
% analysis images in a folder, let's say, ./outputims/:
%     rigor_batch_run(1, 1, '~/pascalvoc12/JPEGImages', ...
%                     '~/pascalvoc12/SegmentationObject', ...
%                     '~/pascalvoc12/ImageSets/Segmentation/val.txt', ...
%                     '.jpg', 'outputims', true, true, 'pascalvoc', ...
%                     'force_recompute', true);
%
% @output:
% The main output is to disk. It generates multiple files, one for each 
% input image, containing object proposals for it. If gt_dir is not empty, 
% it also outputs one score file storing the performance compared to the 
% GT. For more details see ~/rigor_src/rigor_obj_segments.m
%
% @input:
%   SECTIONS: number of sections to divide the image set into. This is 
%       useful for dividing computation over different machines. See 
%       section_no for more info.
%
%   section_no: this number should be in the range [1, SECTIONS], and 
%       specifies which subset of images this function call has to compute. 
%       This is useful when dividing computation over different machines. 
%       If SECTIONS is set to 4, and section_no is 3, the function call 
%       would compute object proposals for the 3rd quartile of images.
%
%   data_dir: the directory where the images are stored. The image 
%       filenames (without the extension) that need to be processed are in 
%       the contents of batchlist_filepath
%
%   gt_dir: the directory where GT images are stored. If dataset is Pascal 
%       VOC, then the corresponding GT is found by matching the filename, 
%       and replacing the extension with .png. If dataset is MS COCO, then 
%       this should be the path to the file produced by 
%       pycocotools_cpy_mat.py. If this path/file is provided, it generates 
%       the accuracy scores of all segments against the GT by calling 
%       compute_print_results.m (see the file for details). If gt_dir is an 
%       empty string '', the function call doesn't produce any scores.
%
%   batchlist_filepath: Is the path to a simple text file, where each line 
%       specifies the image's filename (without the extension) that needs 
%       to be processed.
%
%   file_ext: The extension which would be appended to all filenames. Note 
%       that you have to specify the dot with the extension.
%
%   analysis_ims_dir: If do_analysis is true, this is the directory 
%       where the quantitative analysis and timing analysis images are 
%       written to.
%
%   do_analysis: Boolean deciding whether to do any quantitative scoring 
%       and timing analysis by invoking rigor_analysis.m or not.
%
%   per_im_out: This only matters when do_analysis is true, it then forces
%       rigor_analysis to produce an analysis image for each image. In
%       either case it would generate summary images (which account for the
%       whole set of images passed to this function)
%
%   dataset_id: Identifies (by a string) what dataset this is. This is only
%       necessary if you want to get quantitative results against some GT,
%       i.e. when gt_dir is given a valid value. This helps 
%       compute_print_results() so that it knows how to load GT data, and 
%       rigor_analysis() can load the category strings. The two options for
%       now are 'pascalvoc' (PASCAL VOC), 'coco2014' (Microsoft COCO).
%
%   varargin: are the parameters that are directly passed to each call of 
%       rigor_obj_segments.m
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Fall 2014

    if ~exist('analysis_ims_dir', 'var'), analysis_ims_dir = '.'; end
    if ~exist('do_analysis','var'), do_analysis = true; end
    if ~exist('per_im_out','var'), per_im_out = true; end
    if ~exist('dataset_id', 'var') || isempty(dataset_id)
        dataset_id = 'pascalvoc';
    end
    
    % load the filenames
    batch_files = importdata(batchlist_filepath);
    
    % add RIGOR to the path
    rigor_root = fullfile(fileparts(which(mfilename)), '..');
    addpath(rigor_root);
    addpath(fullfile(rigor_root, 'utils'));
    
    % find what files to work on
    if ~exist('SECTIONS', 'var'), SECTIONS = 1; end
    if ~exist('section_no', 'var'), section_no = 1; end
    [start_idx, end_idx] = give_section(length(batch_files), ...
        SECTIONS, section_no);
    
    
    if strcmp(dataset_id, 'coco2014') == 1 && ~isempty(gt_dir)
        % if coco, preload the coco data (gt_dir should be a filepath to
        % the json file)
        % first add coco MATLAB api to the path
        coco_root = fileparts(fileparts(gt_dir));
        addpath(fullfile(coco_root, 'MatlabAPI'));
        % load coco data
        gt_dir = CocoApi(gt_dir);
    end
    
    [~, f] = fileparts(batchlist_filepath);
    fprintf(['RUNNING RIGOR ON %s (%s) SECTION %d out of %d: %d - ', ...
             '%d\n//////////////////////////////////////////////////', ...
             '/////////////////////////\n'], dataset_id, f, section_no, ...
             SECTIONS, start_idx, end_idx);
    
    segobj_filepaths = {};
    scores_filepaths = {};
    preload_data = struct;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % compute RIGOR for all files in the section assigned %
    for file_idx = start_idx:end_idx
        img_name = batch_files{file_idx};
        
        filepath = fullfile(data_dir, [img_name file_ext]);

        [segs, sp_seg, seg_obj] = ...
            rigor_obj_segments(filepath, varargin{:}, ...
                               'preload', preload_data, ...
                               'includes_done', (file_idx > start_idx));
        
        preload_data = seg_obj.preload_data;
        
        % compute scores; print results; save
        if ~isempty(gt_dir)
            segs = convert_masks(sp_seg, segs);
            compute_print_results(segs, seg_obj, gt_dir, dataset_id);
        end
        fprintf('\n');
        
        [~, segobj_filepaths{end+1}] = seg_obj.return_seg_save_filepath();
        scores_filepaths{end+1} = seg_obj.return_scores_filepath();
    end
    
    % analyze the timing info of all the calls to rigor_obj_segments
    if do_analysis
        catg_strs = dataset_catg_strs(dataset_id, gt_dir);
        [avg_timings, avg_stats] = ...
            rigor_analysis(segobj_filepaths, analysis_ims_dir, ...
                           per_im_out, catg_strs);

        disp(avg_stats);
    end
    
    fprintf('\n<<<<<<<<<<<<<<<<<<<<<<< FIN >>>>>>>>>>>>>>>>>>>>>>>\n');
end

function [start_idx, end_idx, section_sz] = give_section(total_len, ...
        SECTIONS, section_no)
    % get which params this program run will compute
    section_sz = ceil(total_len / SECTIONS);
    start_idx = section_sz * (section_no-1) + 1;
    end_idx = section_sz * section_no;
    if end_idx > total_len
        end_idx = total_len;
    end
    section_sz = end_idx - start_idx + 1;
end

