function [Q, collated_scores] = ...
           proposal_quality(img_name, gt_data, segs, segm_quality_type, ...
                            dataset_id, extern_src_dir)
%SEGMENT_QUALITY computes qualitative scores of segment proposals against
% some given GT. Currently this function only runs for PASCAL VOC
% (dataset_id='pascalvoc') or Microsoft COCO (dataset_id='coco2014')
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Spring 2015

    if iscell(segs)
        segs = cell2mat(segs);
    end
    if ndims(segs) == 4
        if size(segs,3) > 1
            warning('foo:myCalcCandScoreFigureGroundAll', ...
                'Looking at only the first image for scoring against GT');
        end
        segs = squeeze(segs(:,:,1,:));
    end
    
    % if just need bounding box overlap scores for PASCAL VOC
    if strcmpi(dataset_id,'pascalvoc') && strcmpi(segm_quality_type, ...
                                                  'bb_overlap')
        [Q, collated_scores] = pascalvoc_bb(img_name, gt_data, segs, ...
                                            extern_src_dir);
        return;
    end

    Q = struct('q',[], 'gt_seg_szs',[], 'gt_bb_szs',[], ...
               'gt_category_id',[], 'best_overlap',[], ...
               'best_seg_idx',[], 'extra_info',struct);
    Q = repmat(Q, [1,0]);
    bb_ranges = [];
    
    % first retrieve the GT data (and info what requires separate logic for
    % each different dataset)
    if strcmpi(dataset_id,'pascalvoc')
        % gt_data is the directory path when pascalvoc
        name = fullfile(gt_data, [img_name '.png']);
        ground_truth_obj_segs = imread(name);

        name = fullfile(gt_data, '../SegmentationClass', [img_name '.png']);
        ground_truth_obj_ctg = imread(name);
        
        un = unique(ground_truth_obj_segs);
        un(un==0) = [];
        un(un==255) = [];
        
        num_gts = numel(un);
        
        % collate all the GT into one 3D matrix, where each pixel is 1 for
        % ignored pixels, and 2 for the actual GT positive
        all_gt = zeros([numel(ground_truth_obj_segs), num_gts]);
        gt_ids = [1 2];
        
        % iterate over each GT and separate otu each GT, compute the bb 
        % range, and get the GT category
        for k = 1:num_gts
            all_gt(ground_truth_obj_segs == 255, k) = gt_ids(1);
            all_gt(ground_truth_obj_segs == un(k), k) = gt_ids(2);
            
            % store the bb range
            [r, c] = find(ground_truth_obj_segs == un(k));
            temp = [min([r, c],[],1), max([r, c],[],1)];
            bb_ranges = [bb_ranges; temp];
            
            % note the segment category type
            Q(k).gt_category_id = ground_truth_obj_ctg(r(1), c(1));
        end
        
        all_gt = reshape(all_gt, [size(ground_truth_obj_segs), num_gts]);
        
        % check if we have category ids for all GTs
        assert(~any(ismember([Q.gt_category_id], [0,255])), ...
               'Couldn''t find the category id of one or more segments');
           
    elseif strcmpi(dataset_id,'coco2014')
        % retrieve image info
        im_id = regexp(img_name, 'COCO_val2014_0*(\d+)', 'tokens');
        assert(numel(im_id) == 1, ...
               'Invalid number of image id regex matches');
        im_id = str2double(im_id{1}{1});
        img_info = gt_data.loadImgs(im_id);
        assert(strfind(img_info.file_name, img_name) == 1, ...
               'Something went wrong while matching image info');
    
        % retrieve the GT for coco
        [all_gt, extra_info] = cocoGenGTMask(gt_data, im_id);
        gt_ids = 1;
        
        num_gts = size(all_gt, 3);

        % iterate over each GT and get the bb range, GT category
        for k = 1:num_gts
            % store the bb range
            bb_ranges = [bb_ranges; extra_info(k).bb_range];

            % note the segment category type
            Q(k).gt_category_id = extra_info(k).gt_category_id;
        end
    else
        error(['Scoring against GT of ''%s'' dataset hasn'' been ' ...
               'implemented yet'], dataset_id);
    end
    
    % get other info about GT (which doesn't need separate logic for each
    % different dataset)
    for k = 1:num_gts
        Q(k).gt_seg_szs = uint32(nnz(all_gt(:,:,k) == gt_ids(end)));
        
        % compute the size of the bb
        temp = bb_ranges(k,:);
        Q(k).gt_bb_szs = uint32(prod(temp([3 4]) - temp([1 2]) + 1));
        % if size out of bounds
        if isinf(Q(k).gt_bb_szs) || Q(k).gt_bb_szs < 0
            Q(k).gt_bb_szs = 0;
        end

        % default values (which would be values when no segs or no GT)
        Q(k).q = zeros(0,1,'single');
        
        Q(k).best_overlap = single(0);
        Q(k).best_seg_idx = int32(-1);
        Q(k).extra_info.overlap_bb = single(0);
        Q(k).extra_info.best_bb_overlap = single(0);
        Q(k).extra_info.best_bb_seg_idx = int32(-1);
    end
    
    % only compute scores if there are non-zero # of segments
    if size(segs,3) > 0
        % if overlap compute using the faster mex code
        if strcmpi(segm_quality_type, 'overlap')
            ign_gt = [];
            if numel(gt_ids) == 2
                ign_gt = all_gt == gt_ids(1);
            end
            % run fast overlap computation (with ignrd pixels if available)
            [this_Q, obb] = fast_segm_overlap_care_bb_mex(segs, ...
                                          all_gt == gt_ids(end), ign_gt);
            
            [best_bb_overlap, best_bb_idx] = max(obb,[],1);
            best_bb_idx = int32(best_bb_idx);
            for k = 1:num_gts
                Q(k).extra_info.overlap_bb = obb(:,k);
                Q(k).extra_info.best_bb_overlap = best_bb_overlap(k);
                Q(k).extra_info.best_bb_seg_idx = best_bb_idx(k);
            end
        else
            % if not overlap, use the computer_error_metric function
            this_Q = zeros(size(segs,3), num_gts, 'single');
            parfor k = 1:num_gts
                [q, ei] = compute_error_metric(segs, all_gt(:,:,k), ...
                                               gt_ids, segm_quality_type);
                this_Q(:,k) = q;
                Q(k).extra_info = ei;
            end
        end

        % collate results for all 
        for k = 1:num_gts
            Q(k).q = this_Q(:,k);

            % find the best segment
            [best_overlap, best_seg_idx] = max(Q(k).q);
            Q(k).best_overlap = best_overlap;
            Q(k).best_seg_idx = int32(best_seg_idx);
        end
    end
    
    % collate the scores
    collated_scores.avg_best_overlap = mean([Q.best_overlap]);
    collated_scores.covering = sum([Q.best_overlap] .* ...
                             double([Q.gt_seg_szs])) / sum([Q.gt_seg_szs]);
end


function [Q, collated_scores] = pascalvoc_bb(img_name, gt_dir, segs, ...
                                             extern_src_dir)
    % NOTE THIS CURRENTLY ONLY WORKS WITH VOC_STYLE ANNOTATIONS

    % add PASCAL VOC code for running PASreadrecord
    addpath(fullfile(extern_src_dir, ...
                     'toolboxes/everingham_IJCV_2010_pascalvoc/VOCcode/'));
    % get the VOC xml annotations
    xml_filepath = fullfile(gt_dir, '../Annotations', [img_name '.xml']);
    rec = PASreadrecord(xml_filepath);
    to_remove = [rec.objects.difficult];
    rec.objects = rec.objects(~to_remove);
    [~,all_bbox] = segment_regionprops(segs);
    all_bbox = single(all_bbox);
    for k=1:numel(rec.objects)
       box_max = bsxfun(@max, all_bbox, rec.objects(k).bbox);
       box_min = bsxfun(@min, all_bbox, rec.objects(k).bbox);
       box_int = max((box_min(:,3) - box_max(:,1)+1),0) .* ...
                 max((box_min(:,4) - box_max(:,2)+1),0);
       box_uni = max((box_max(:,3) - box_min(:,1)+1),0) .* ...
                 max((box_max(:,4) - box_min(:,2)+1),0);
       Q(k).q = box_int ./ box_uni;
       Q(k).gt_seg_szs = ...
           (rec.objects(k).bbox(4) - rec.objects(k).bbox(2)+1) * ...
           (rec.objects(k).bbox(3) - rec.objects(k).bbox(1)+1);
       [Q(k).best_overlap, Q(k).best_seg_idx] = max(Q(k).q);
       Q(k).extra_info.tp = box_int;
       Q(k).extra_info.fp = (all_bbox(:,4) - all_bbox(:,2) + 1) .* ...
                            (all_bbox(:,3) - all_bbox(:,1) + 1) ...
                            - Q(k).extra_info.tp;
       Q(k).extra_info.fn = Q(k).gt_seg_szs - Q(k).extra_info.tp;
    end
%     un = numel(rec.objects);
    
    % collate the scores
    collated_scores.avg_best_overlap = mean([Q.best_overlap]);
    collated_scores.covering = sum([Q.best_overlap] .* ...
                             double([Q.gt_seg_szs])) / sum([Q.gt_seg_szs]);
end